# Bob build tool
# Copyright (C) 2016  TechniSat Digital GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later

from . import BOB_VERSION, _enableDebug, DEBUG
from .errors import BobError
from .state import finalize
from .tty import colorize, Unbuffered, setColorMode, cleanup
from .utils import asHexStr, hashPath, getPlatformTag, EventLoopWrapper
import argparse
import logging
import sys
import traceback
import os

def __archive(*args, **kwargs):
     from .cmds.archive import doArchive
     doArchive(*args, **kwargs)
     return 0

def __build(*args, **kwargs):
     from .cmds.build.build import doBuild
     doBuild(*args, **kwargs)
     return 0

def __develop(*args, **kwargs):
     from .cmds.build.build import doDevelop
     doDevelop(*args, **kwargs)
     return 0

def __clean(*args, **kwargs):
     from .cmds.build.clean import doClean
     doClean(*args, **kwargs)
     return 0

def __graph(*args, **kwargs):
    from .cmds.graph import doGraph
    doGraph(*args, **kwargs)
    return 0

def __help(*args, **kwargs):
    from .cmds.help import doHelp
    doHelp(availableCommands.keys(), *args, **kwargs)
    return 0

def __jenkins(*args, **kwargs):
     from .cmds.jenkins import doJenkins
     doJenkins(*args, **kwargs)
     return 0

def __ls(*args, **kwargs):
     from .cmds.misc import doLS
     doLS(*args, **kwargs)
     return 0

def __project(*args, **kwargs):
     from .cmds.build.project import doProject
     doProject(*args, **kwargs)
     return 0

def __status(*args, **kwars):
     from .cmds.build.status import doStatus
     doStatus(*args, **kwars)
     return 0

def __queryscm(*args, **kwargs):
     from .cmds.misc import doQuerySCM
     doQuerySCM(*args, **kwargs)
     return 0

def __querymeta(*args, **kwargs):
     from .cmds.misc import doQueryMeta
     doQueryMeta(*args, **kwargs)
     return 0

def __queryrecipe(*args, **kwargs):
     from .cmds.misc import doQueryRecipe
     doQueryRecipe(*args, **kwargs)
     return 0

def __querypath(*args, **kwargs):
     from .cmds.build.query import doQueryPath
     doQueryPath(*args, **kwargs)
     return 0

def __download(*args, **kwargs):
    from .archive import doDownload
    doDownload(*args, **kwargs)
    return 0

def __upload(*args, **kwargs):
    from .archive import doUpload
    doUpload(*args, **kwargs)
    return 0

def __invoke(*args, **kwargs):
    from .cmds.invoke import doInvoke
    return doInvoke(*args, **kwargs)

availableCommands = {
    "archive"       : ('hl', __archive, "Manage binary artifact archives"),
    "build"         : ('hl', __build, "Build (sub-)packages in release mode"),
    "dev"           : ('hl', __develop, "Build (sub-)packages in development mode"),
    "clean"         : ('hl', __clean, "Delete unused src/build/dist paths of release builds"),
    "graph"         : ('hl', __graph, "Make a interactive dependency graph"),
    "help"          : ('hl', __help, "Display help information about command"),
    "jenkins"       : ('hl', __jenkins, "Configure Jenkins server"),
    "ls"            : ('hl', __ls, "List package hierarchy"),
    "project"       : ('hl', __project, "Create project files"),
    "status"        : ('hl', __status, "Show SCM status"),

    "query-scm"     : ('ll', __queryscm, "Query SCM information"),
    "query-recipe"  : ('ll', __queryrecipe, "Query package sources"),
    "query-path"    : ('ll', __querypath, "Query path information"),
    "query-meta"    : ('ll', __querymeta, "Query Package meta information"),

    "_download"     : (None, __download, ""),
    "_upload"       : (None, __upload, ""),
    "_invoke"       : (None, __invoke, ""),
}

def describeCommands():
    hlCmds = "\n".join(sorted([ "  {:16s}{}".format(k, v[2])
        for (k,v) in availableCommands.items() if v[0] == 'hl' ]))
    llCmds = "\n".join(sorted([ "  {:16s}{}".format(k, v[2])
        for (k,v) in availableCommands.items() if v[0] == 'll' ]))
    return """
The following high level commands are available:

{}

The following scripting commands are available:

{}

See 'bob <command> -h' for more information on a specific command.""".format(hlCmds, llCmds);

def catchErrors(fun, *args, **kwargs):
    try:
        ret = fun(*args, **kwargs)
    except BrokenPipeError:
        # explicitly close stderr to suppress further error messages
        sys.stderr.close()
        ret = 0
    except BobError as e:
        print(e, file=sys.stderr)
        ret = e.returncode
    except KeyboardInterrupt:
        ret = 2
    except ImportError as e:
        if e.name:
            print(colorize("Python module '{}' seems to be missing. ".format(e.name) +
                           "Please check your installation...", "31;1"),
                  file=sys.stderr)
        else:
            print(colorize(str(e) + " Please check your installation...", "31;1"),
                  file=sys.stderr)
        ret = 3
    except Exception:
        print(colorize("""An internal Exception has occured. This should not have happenend.
Please open an issue at https://github.com/BobBuildTool/bob with the following backtrace:""", "31;1"), file=sys.stderr)
        print("Bob version", BOB_VERSION, file=sys.stderr)
        traceback.print_exc(file=sys.stderr)
        ret = 3

    return ret

def bob(bobRoot = None):
    if not bobRoot:
        bobRoot = os.path.realpath(os.path.abspath(sys.argv[0]))
    origSysStdOut = sys.stdout
    origSysStdErr = sys.stderr
    logging.disable(logging.ERROR)

    # Prevent any buffering. Even on a tty Python is doing line buffering.
    sys.stdout = Unbuffered(sys.stdout)
    sys.stderr = Unbuffered(sys.stderr)

    def cmd():
        parser = argparse.ArgumentParser(prog="bob",
                                         description="Bob build tool\n" + describeCommands(),
                                         formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument('-C', '--directory', dest='directory', action='append', help="Change to DIRECTORY before doing anything", metavar="DIRECTORY")
        parser.add_argument('--version', dest='version', action='store_true', help="Show version")
        parser.add_argument('--debug',   dest='debug', help="Enable debug modes (shl,pkgck,ngd)")
        parser.add_argument('-i', dest='ignore_commandCfg', default=False, action='store_true',
                help="Use bob's default argument settings and do not use commands section of the userconfig.")
        parser.add_argument('--color', dest='color_mode',
                help="Color mode of console output (default: auto)",
                choices=['never', 'always', 'auto'])
        parser.add_argument('command', nargs='?', help="Command to execute")
        parser.add_argument('args', nargs=argparse.REMAINDER, help="Arguments to command")

        args = parser.parse_args(sys.argv[1:])
        if args.version:
            print("Bob version", BOB_VERSION)
            return 0

        if args.debug:
            _enableDebug(args.debug)

        if args.ignore_commandCfg:
            from .input import RecipeSet
            RecipeSet.ignoreCommandCfg()

        if args.color_mode:
            from .input import RecipeSet
            RecipeSet.setColorModeCfg(args.color_mode)

        if args.command is None:
            print("No command specified. Use 'bob -h' for help.", file=sys.stderr)
            return 2

        # Shortcut for 'bob help' displaying the same help screen like 'bob
        # --help' would do. The 'bob help COMMAND' case is handled by help from
        # the available commands.
        if args.command == "help" and len(args.args) == 0:
            parser.print_help()
            return 0

        if args.command in availableCommands:
            if args.directory is not None:
                for i in args.directory:
                    try:
                        os.chdir(i)
                    except OSError as e:
                        print("bob -C: unable to change directory:", str(e), file=sys.stderr)
                        return 1
            cmd = availableCommands[args.command][1]
            if DEBUG['prof']:
                import cProfile, pstats
                pr = cProfile.Profile()
                pr.enable()
                ret = cmd(args.args, bobRoot)
                pr.disable()
                ps = pstats.Stats(pr, stream=sys.stderr).sort_stats('tottime')
                print("Bob", BOB_VERSION, "profile:", file=sys.stderr)
                print("Args:", sys.argv[1:], file=sys.stderr)
                ps.print_stats()
            else:
                ret = cmd(args.args, bobRoot)
            return ret
        else:
            print("Don't know what to do for '{}'. Use 'bob -h' for help".format(args.command), file=sys.stderr)
            return 2

    try:
        ret = catchErrors(cmd)
    finally:
        cleanup()
        sys.stdout = origSysStdOut
        sys.stderr = origSysStdErr
        finalize()

    return ret

def hashEngine():
    parser = argparse.ArgumentParser(description="Create hash based on spec.")
    parser.add_argument('-o', dest="output", metavar="OUTPUT", default="-", help="Output file (default: stdout)")
    parser.add_argument('--state', help="State cache directory")
    parser.add_argument('spec', nargs='?', default="-", help="Spec input (default: stdin)")
    args = parser.parse_args()

    def cmd():
        try:
            if args.spec == "-":
                inFile = sys.stdin
            else:
                inFile = open(args.spec, "r")

            res = b''
            for cmd in inFile:
                res += __process(cmd.strip(), inFile, args.state)
            if args.output == "-":
                sys.stdout.buffer.write(res)
            else:
                with open(args.output, "wb") as f:
                    f.write(res)
        except OSError as e:
            raise BobError("IO error: " + str(e))

        return 0

    return catchErrors(cmd)

def auditEngine():
    parser = argparse.ArgumentParser(description="Create audit trail.")
    parser.add_argument('-o', dest="output", metavar="OUTPUT", default="-", help="Output file (default: stdout)")
    parser.add_argument("-D", dest="defines", action="append", default=[], nargs=2)
    parser.add_argument("--arg", action="append", default=[])
    parser.add_argument("--env")
    parser.add_argument("-E", dest="metaEnv", action="append", default=[], nargs=2)
    parser.add_argument("--recipes")
    parser.add_argument("--sandbox")
    parser.add_argument("--scm", action="append", default=[], nargs=3) # legacy Bob <= 0.15
    parser.add_argument("--scmEx", action="append", default=[], nargs=4)
    parser.add_argument("--tool", action="append", default=[], nargs=2)
    parser.add_argument("variantID")
    parser.add_argument("buildID")
    parser.add_argument("resultHash")
    args = parser.parse_args()

    def cmd(loop):
        from .audit import Audit
        import json
        import os
        try:
            gen = Audit.create(bytes.fromhex(args.variantID), bytes.fromhex(args.buildID),
                bytes.fromhex(args.resultHash))
        except ValueError:
            raise BuildError("Invalid digest argument")
        if args.env is not None: gen.setEnv(args.env)
        for (name, value) in args.metaEnv: gen.addMetaEnv(name, value)
        for (name, value) in args.defines: gen.addDefine(name, value)
        for (name, workspace, dir) in args.scm:
            loop.run_until_complete(gen.addScm(name, workspace, dir, {}))
        for (name, workspace, dir, extra) in args.scmEx:
            loop.run_until_complete(gen.addScm(name, workspace, dir,
                json.loads(extra)))
        for (name, audit) in args.tool: gen.addTool(name, audit)
        try:
            if args.recipes is not None: gen.setRecipesData(json.loads(args.recipes))
        except ValueError as e:
            raise BuildError("Invalid recipes json: " + str(e))
        if args.sandbox is not None: gen.setSandbox(args.sandbox)
        for arg in args.arg: gen.addArg(arg)

        if args.output == "-":
            gen.save(sys.stdout.buffer)
        else:
            gen.save(args.output)

        return 0

    with EventLoopWrapper() as loop:
        return catchErrors(cmd, loop)

def __process(l, inFile, stateDir):
    if l.startswith("="):
        return bytes.fromhex(l[1:])
    elif l.startswith("<"):
        with open(l[1:], "rb") as f:
            return f.read()
    elif l.startswith("{"):
        import hashlib
        fn = l[1:]
        skipEmpty = False
        if fn.startswith("?"):
            fn = fn[1:]
            skipEmpty = True
        return __processBlock(hashlib.new(fn), inFile, stateDir, skipEmpty)
    elif l.startswith("["):
        exp,sep,l = l[1:].partition("]")
        start,sep,end = exp.partition(":")
        start = int(start) if start else None
        end = int(end) if end else None
        return __process(l, inFile, stateDir)[start:end]
    elif l.startswith("#"):
        import os.path
        if stateDir:
            stateFile = os.path.join(stateDir, l[1:].replace(os.sep, "_"))
        else:
            stateFile = None
        return hashPath(l[1:], stateFile)
    elif l.startswith("g"):
        from .scm.git import GitScm
        return bytes.fromhex(GitScm.processLiveBuildIdSpec(l[1:]))
    elif l == "p":
        return getPlatformTag()
    elif l:
        print("Malformed spec:", l, file=sys.stderr)
        sys.exit(1)
    else:
        return b''

def __processBlock(h, inFile, stateDir, skipEmpty):
    for l in inFile:
        l = l.strip()
        if l.startswith("}"):
            return h.digest() if not skipEmpty else b''
        else:
            data = __process(l, inFile, stateDir)
            if data: skipEmpty = False
            h.update(data)
    print("Malformed spec: unfinished block", file=sys.stderr)
    sys.exit(1)

if __name__ == '__main__':
    if sys.argv[1] == 'bob':
        rootDir = sys.argv[2]
        del sys.argv[1:3]
        sys.exit(bob(rootDir))




# Bob build tool
# Copyright (C) 2016-2018  TechniSat Digital GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ...errors import BobError
from ...generators import generators as defaultGenerators
from ...input import RecipeSet
from ...tty import colorize
from ...utils import processDefines
import argparse
import os

from .build import doDevelop
from .builder import LocalBuilder
from .state import DevelopDirOracle

def doProject(argv, bobRoot):
    def _downloadArg(arg):
        if (arg.startswith("packages=") or arg in ['yes', 'no', 'deps']):
            return arg
        raise argparse.ArgumentTypeError("{} invalid.".format(arg))

    parser = argparse.ArgumentParser(prog="bob project", description='Generate Project Files')
    parser.add_argument('projectGenerator', nargs='?', help="Generator to use.")
    parser.add_argument('package', nargs='?', help="Sub-package that is the root of the project")
    parser.add_argument('args', nargs=argparse.REMAINDER,
                        help="Arguments for project generator")

    parser.add_argument('--list', default=False, action='store_true', help="List available Generators")
    parser.add_argument('-D', default=[], action='append', dest="defines",
        help="Override default environment variable")
    parser.add_argument('-c', dest="configFile", default=[], action='append',
        help="Use config File")
    parser.add_argument('-e', dest="white_list", default=[], action='append', metavar="NAME",
        help="Preserve environment variable")
    parser.add_argument('-E', dest="preserve_env", default=False, action='store_true',
        help="Preserve whole environment")
    parser.add_argument('--download', metavar="MODE", default="no",
        help="Download from binary archive (yes, no, deps, packages=<regular expression>)",
        type=_downloadArg)
    parser.add_argument('--resume', default=False, action='store_true',
        help="Resume build where it was previously interrupted")
    parser.add_argument('-n', dest="execute_prebuild", default=True, action='store_false',
        help="Do not build (bob dev) before generate project Files. RunTargets may not work")
    parser.add_argument('-b', dest="execute_buildonly", default=False, action='store_true',
        help="Do build only (bob dev -b) before generate project Files. No checkout")
    parser.add_argument('-j', '--jobs', default=None, type=int, nargs='?', const=...,
        help="Specifies  the  number of jobs to run simultaneously.")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--sandbox', action='store_true', default=False,
        help="Enable sandboxing")
    group.add_argument('--no-sandbox', action='store_false', dest='sandbox',
        help="Disable sandboxing")
    args = parser.parse_args(argv)

    defines = processDefines(args.defines)

    recipes = RecipeSet()
    recipes.defineHook('developNameFormatter', LocalBuilder.developNameFormatter)
    recipes.defineHook('developNamePersister', None)
    recipes.setConfigFiles(args.configFile)
    recipes.parse()

    envWhiteList = recipes.envWhiteList()
    envWhiteList |= set(args.white_list)

    nameFormatter = recipes.getHook('developNameFormatter')
    developPersister = DevelopDirOracle(nameFormatter, recipes.getHook('developNamePersister'))
    nameFormatter = LocalBuilder.makeRunnable(developPersister.getFormatter())
    packages = recipes.generatePackages(nameFormatter, defines, sandboxEnabled=args.sandbox)
    developPersister.prime(packages)

    generators = defaultGenerators.copy()
    generators.update(recipes.getProjectGenerators())

    if args.list:
        for g in generators:
            print(g)
        return 0
    else:
        if not args.package or not args.projectGenerator:
            raise BobError("The following arguments are required: projectGenerator, package")

    try:
        generator = generators[args.projectGenerator]
    except KeyError:
        raise BobError("Generator '{}' not found!".format(args.projectGenerator))

    extra = [ "--download=" + args.download ]
    for d in args.defines:
        extra.append('-D')
        extra.append(d)
    for c in args.configFile:
        extra.append('-c')
        extra.append(c)
    for e in args.white_list:
        extra.append('-e')
        extra.append(e)
    if args.preserve_env: extra.append('-E')
    if args.sandbox: extra.append('--sandbox')
    if args.jobs is ...:
        # expand because we cannot control the argument order in the generator
        args.jobs = os.cpu_count()
    if args.jobs is not None:
        if args.jobs <= 0:
            parser.error("--jobs argument must be greater than zero!")
        extra.extend(['-j', str(args.jobs)])

    package = packages.walkPackagePath(args.package)

    # execute a bob dev with the extra arguments to build all executables.
    # This makes it possible for the plugin to collect them and generate some runTargets.
    if args.execute_prebuild:
        devArgs = extra.copy()
        if args.resume: devArgs.append('--resume')
        if args.execute_buildonly: devArgs.append('-b')
        devArgs.append(args.package)
        doDevelop(devArgs, bobRoot)

    print(">>", colorize("/".join(package.getStack()), "32;1"))
    print(colorize("   PROJECT   {} ({})".format(args.package, args.projectGenerator), "32"))
    generator(package, args.args, extra, bobRoot)




# Bob build tool
# Copyright (C) 2019  Jan Klötzke
#
# SPDX-License-Identifier: GPL-3.0-or-later

from .common import CommonIDEGenerator
from ..utils import quoteCmdExe
from pathlib import Path, PureWindowsPath
from shlex import quote as quoteBash
from uuid import UUID
from uuid import uuid4 as randomUuid
from uuid import uuid5 as sha1NsUuid
from xml.etree import ElementTree
import os
import sys

SOLUTION_TEMPLATE = """\

Microsoft Visual Studio Solution File, Format Version 12.00
# Visual Studio Version 16
VisualStudioVersion = 16.0.28803.352
MinimumVisualStudioVersion = 10.0.40219.1
{PROJECTS_LIST}
Global
	GlobalSection(SolutionConfigurationPlatforms) = preSolution
		Build|x86 = Build|x86
		Checkout+Build|x86 = Checkout+Build|x86
	EndGlobalSection
	GlobalSection(ProjectConfigurationPlatforms) = postSolution
{PROJECTS_CFG}
	EndGlobalSection
	GlobalSection(SolutionProperties) = preSolution
		HideSolutionNode = FALSE
	EndGlobalSection
	GlobalSection(ExtensibilityGlobals) = postSolution
		SolutionGuid = {{{SOLUTION_GUID}}}
	EndGlobalSection
EndGlobal
"""

SOLUTION_PROJECT_TEMPLATE = """\
Project("{{8BC9CEB8-8B4A-11D0-8D11-00A0C91BC942}}") = "{NAME}", "{NAME}\\{NAME}.vcxproj", "{{{GUID}}}"
EndProject"""

PROJECT_TEMPLATE = """\
<?xml version="1.0" encoding="utf-8"?>
<Project DefaultTargets="Build" xmlns="http://schemas.microsoft.com/developer/msbuild/2003">
  <ItemGroup Label="ProjectConfigurations">
    <ProjectConfiguration Include="Build|Win32">
      <Configuration>Build</Configuration>
      <Platform>Win32</Platform>
    </ProjectConfiguration>
    <ProjectConfiguration Include="Checkout+Build|Win32">
      <Configuration>Checkout+Build</Configuration>
      <Platform>Win32</Platform>
    </ProjectConfiguration>
  </ItemGroup>
  <PropertyGroup Label="Globals">
    <VCProjectVersion>16.0</VCProjectVersion>
    <ProjectGuid>{{{PROJECT_GUID}}}</ProjectGuid>
    <Keyword>Win32Proj</Keyword>
  </PropertyGroup>
  <Import Project="$(VCTargetsPath)\Microsoft.Cpp.Default.props" />
  <PropertyGroup Condition="'$(Configuration)|$(Platform)'=='Build|Win32'" Label="Configuration">
    <ConfigurationType>Makefile</ConfigurationType>
    <UseDebugLibraries>true</UseDebugLibraries>
    <PlatformToolset>v142</PlatformToolset>
  </PropertyGroup>
  <PropertyGroup Condition="'$(Configuration)|$(Platform)'=='Checkout+Build|Win32'" Label="Configuration">
    <ConfigurationType>Makefile</ConfigurationType>
    <UseDebugLibraries>true</UseDebugLibraries>
    <PlatformToolset>v142</PlatformToolset>
  </PropertyGroup>
  <Import Project="$(VCTargetsPath)\Microsoft.Cpp.props" />
  <ImportGroup Label="ExtensionSettings">
  </ImportGroup>
  <ImportGroup Label="Shared">
  </ImportGroup>
  <ImportGroup Label="PropertySheets" Condition="'$(Configuration)|$(Platform)'=='Build|Win32'">
    <Import Project="$(UserRootDir)\Microsoft.Cpp.$(Platform).user.props" Condition="exists('$(UserRootDir)\Microsoft.Cpp.$(Platform).user.props')" Label="LocalAppDataPlatform" />
  </ImportGroup>
  <ImportGroup Label="PropertySheets" Condition="'$(Configuration)|$(Platform)'=='Checkout+Build|Win32'">
    <Import Project="$(UserRootDir)\Microsoft.Cpp.$(Platform).user.props" Condition="exists('$(UserRootDir)\Microsoft.Cpp.$(Platform).user.props')" Label="LocalAppDataPlatform" />
  </ImportGroup>
  <PropertyGroup Label="UserMacros" />
  <ItemDefinitionGroup>
  </ItemDefinitionGroup>
  <Import Project="$(VCTargetsPath)\Microsoft.Cpp.targets" />
  <ImportGroup Label="ExtensionTargets">
  </ImportGroup>
{EXTENSIONS}
</Project>
"""

FILTERS_SOURCES_UUID = UUID("4FC737F1-C7A5-4376-A066-2A32D752A2FF")
FILTERS_HEADERS_UUID = UUID("93995380-89BD-4b04-88EB-625FBE52EBFB")
FILTERS_RESOURCES_UUID = UUID("67DA6AB6-F800-4c08-8B7A-83BB121AAD01")

FILTERS_TEMPLATE = """\
<?xml version="1.0" encoding="utf-8"?>
<Project ToolsVersion="4.0" xmlns="http://schemas.microsoft.com/developer/msbuild/2003">
  <ItemGroup>
    <Filter Include="Source Files">
      <UniqueIdentifier>{{4FC737F1-C7A5-4376-A066-2A32D752A2FF}}</UniqueIdentifier>
      <Extensions>cpp;c;cxx;def;odl;idl;hpj;bat;asm;asmx</Extensions>
    </Filter>
    <Filter Include="Header Files">
      <UniqueIdentifier>{{93995380-89BD-4b04-88EB-625FBE52EBFB}}</UniqueIdentifier>
      <Extensions>h;hpp;hxx;hm;inl;inc;xsd</Extensions>
    </Filter>
    <Filter Include="Resource Files">
      <UniqueIdentifier>{{67DA6AB6-F800-4c08-8B7A-83BB121AAD01}}</UniqueIdentifier>
      <Extensions>rc;ico;cur;bmp;dlg;rc2;rct;bin;rgs;gif;jpg;jpeg;jpe;resx</Extensions>
    </Filter>
{FILTERS}
  </ItemGroup>
{ITEMS}
</Project>
"""

USER_TEMPLATE = """\
<?xml version="1.0" encoding="utf-8"?>
<Project ToolsVersion="Current" xmlns="http://schemas.microsoft.com/developer/msbuild/2003">
{GROUPS}
</Project>
"""

#def relpath(pathFrom, pathTo):
#    if pathTo.is_absolute():
#        return str(pathTo)
#
#    if pathFrom.is_absolute():
#        ret = [ pathFrom.drive ]
#        pathFrom = list(pathFrom.parts)[1:]
#    else:
#        ret = []
#        pathFrom = list(pathFrom.parts)
#    pathTo = list(pathTo.parts)
#
#    while pathFrom and not 
#
#    return 

class Project:
    def __init__(self, recipesRoot, uuid, scan):
        self.uuid = uuid
        self.isRoot = scan.isRoot
        self.packagePath = scan.stack
        self.workspacePath = recipesRoot.joinpath(PureWindowsPath(scan.workspacePath))
        self.headers =   [ recipesRoot.joinpath(PureWindowsPath(i)) for i in scan.headers   ]
        self.sources =   [ recipesRoot.joinpath(PureWindowsPath(i)) for i in scan.sources   ]
        self.resources = [ recipesRoot.joinpath(PureWindowsPath(i)) for i in scan.resources ]
        self.incPaths =  [ recipesRoot.joinpath(PureWindowsPath(i)) for i in scan.incPaths  ]
        self.dependencies = scan.dependencies
        self.runTargets = [ recipesRoot.joinpath(PureWindowsPath(i)) for i in scan.runTargets ]

    def generateProject(self, projects, cmd):
        items = []
        incPaths = []
        if self.incPaths:
            for i in self.incPaths: incPaths.append(str(i))
        if self.headers:
            g = ElementTree.Element("ItemGroup")
            for i in self.headers: ElementTree.SubElement(g, "ClInclude", {"Include" : str(i)})
            items.append(ElementTree.tostring(g, encoding="unicode"))
        if self.sources:
            g = ElementTree.Element("ItemGroup")
            for i in self.sources: ElementTree.SubElement(g, "ClCompile", {"Include" : str(i)})
            items.append(ElementTree.tostring(g, encoding="unicode"))
        if self.resources:
            g = ElementTree.ElementTree("ItemGroup")
            for i in self.resources: ElementTree.SubElement(g, "Text", {"Include" : str(i)})
            items.append(ElementTree.tostring(g, encoding="unicode"))
        if self.dependencies:
            g = ElementTree.Element("ItemGroup")
            for i in sorted(self.dependencies):
                dep = ElementTree.SubElement(g, "ProjectReference",
                        {"Include" : "..\\{NAME}\\{NAME}.vcxproj".format(NAME=i)})
                ElementTree.SubElement(dep, "Project").text = "{" + str(projects[i].uuid) + "}"
                ElementTree.SubElement(dep, "LinkLibraryDependencies").text = "true"
                incPaths.extend(str(i) for i in projects[i].incPaths)

            items.append(ElementTree.tostring(g, encoding="unicode"))

        n = ElementTree.Element("PropertyGroup", {"Condition" : "'$(Configuration)|$(Platform)'=='Build|Win32'"})
        ElementTree.SubElement(n, "NMakeOutput")
        ElementTree.SubElement(n, "NMakeIncludeSearchPath").text = ";".join(incPaths)
        ElementTree.SubElement(n, "NMakeBuildCommandLine").text = cmd + " -v -b " + self.packagePath
        ElementTree.SubElement(n, "NMakeReBuildCommandLine").text = cmd + " -v -b -f --clean " + self.packagePath
        # TODO: NMakeCleanCommandLine == "rm -rf dev/{build,dist}"?
        items.append(ElementTree.tostring(n, encoding="unicode"))

        n = ElementTree.Element("PropertyGroup", {"Condition" : "'$(Configuration)|$(Platform)'=='Checkout+Build|Win32'"})
        ElementTree.SubElement(n, "NMakeOutput")
        ElementTree.SubElement(n, "NMakeIncludeSearchPath").text = ";".join(incPaths)
        ElementTree.SubElement(n, "NMakeBuildCommandLine").text = cmd + " -v " + self.packagePath
        ElementTree.SubElement(n, "NMakeReBuildCommandLine").text = cmd + " -v -f --clean " + self.packagePath
        # TODO: NMakeCleanCommandLine == "rm -rf dev/{build,dist}"?
        items.append(ElementTree.tostring(n, encoding="unicode"))

        return PROJECT_TEMPLATE.format(PROJECT_GUID=str(self.uuid),
                EXTENSIONS="\n".join(items))

    def generateFilters(self):
        items = []
        filters = {}

        def makeFilter(category, uuid, filePath):
            trail = [ category ]
            for i in filePath.relative_to(self.workspacePath).parent.parts:
                uuid = sha1NsUuid(uuid, i)
                trail.append(i)
                name = "\\".join(trail)
                if name not in filters:
                    f = ElementTree.Element("Filter", {"Include" : name})
                    ElementTree.SubElement(f, "UniqueIdentifier").text = "{" + str(uuid) + "}"
                    filters[name] = ElementTree.tostring(f, encoding="unicode")

            return "\\".join(trail)

        if self.headers:
            g = ElementTree.Element("ItemGroup")
            for i in self.headers:
                s = ElementTree.SubElement(g, "ClInclude", {"Include" : str(i)})
                ElementTree.SubElement(s, "Filter").text = makeFilter("Header Files", FILTERS_HEADERS_UUID, i)
            items.append(ElementTree.tostring(g, encoding="unicode"))
        if self.sources:
            g = ElementTree.Element("ItemGroup")
            for i in self.sources:
                s = ElementTree.SubElement(g, "ClCompile", {"Include" : str(i)})
                ElementTree.SubElement(s, "Filter").text = makeFilter("Source Files", FILTERS_SOURCES_UUID, i)
            items.append(ElementTree.tostring(g, encoding="unicode"))
        if self.resources:
            g = ElementTree.Element("ItemGroup")
            for i in self.resources:
                s = ElementTree.SubElement(g, "Text", {"Include" : str(i)})
                ElementTree.SubElement(s, "Filter").text = makeFilter("Resource Files", FILTERS_RESOURCES_UUID, i)
            items.append(ElementTree.tostring(g, encoding="unicode"))

        return FILTERS_TEMPLATE.format(FILTERS="\n".join(f for n,f in sorted(filters.items())),
                                       ITEMS="\n".join(items))

    def generateUser(self):
        groups = []
        if self.runTargets:
            # VS can handle only one target per project
            target = self.runTargets[0]
            g = ElementTree.Element("PropertyGroup", {"Condition" : "'$(Configuration)|$(Platform)'=='Build|Win32'"})
            ElementTree.SubElement(g, "LocalDebuggerCommand").text = str(target)
            ElementTree.SubElement(g, "DebuggerFlavor").text = "WindowsLocalDebugger"
            groups.append(ElementTree.tostring(g, encoding="unicode"))

            n = ElementTree.Element("PropertyGroup", {"Condition" : "'$(Configuration)|$(Platform)'=='Checkout+Build|Win32'"})
            ElementTree.SubElement(g, "LocalDebuggerCommand").text = str(target)
            ElementTree.SubElement(g, "DebuggerFlavor").text = "WindowsLocalDebugger"
            groups.append(ElementTree.tostring(g, encoding="unicode"))
            # FIXME: print warning when more than one taget exists

        return USER_TEMPLATE.format(GROUPS="\n".join(groups))


class Vs2019Generator(CommonIDEGenerator):
    def __init__(self):
        super().__init__("vs2019", "Generate Visual Studio 2019 solution")
        self.parser.add_argument('--uuid', help="Set solution UUID")

    def configure(self, package, argv):
        super().configure(package, argv)
        if self.args.uuid:
            self.uuid = UUID(self.args.uuid)
        else:
            self.uuid = randomUuid()

    def generate(self, extra, bobRoot):
        super().generate()

        # gather root paths
        bobPwd = Path(os.getcwd())
        if sys.platform == 'msys':
            if os.getenv('WD') is None:
                raise BuildError("Cannot create Visual Studio project for Windows! MSYS2 must be started by msys2_shell.cmd script!")
            msysRoot = PureWindowsPath(os.getenv('WD')) / '..' / '..'
            winPwd = PureWindowsPath(os.popen('pwd -W').read().strip())
            winDestination = PureWindowsPath(os.popen('cygpath -w {}'.format(quoteBash(self.destination))).read().strip())
            baseBuildMe = str(msysRoot / "msys2_shell.cmd") + \
                    " -msys2 -defterm -no-start -use-full-path -where " + \
                    str(winPwd)
            buildMe = os.path.join(self.destination, "buildme.sh")
            buildMeCmd = baseBuildMe + " " + buildMe
        else:
            winPwd = bobPwd
            winDestination = Path(self.destination).resolve()
            buildMe = os.path.join(str(winDestination), "buildme.cmd")
            buildMeCmd = buildMe

        projects = {
            name : Project(winPwd, sha1NsUuid(self.uuid, name), scan)
            for name,scan in self.packages.items()
        }

        if not self.args.update:
            self.updateFile(buildMe, self.__generateBuildme(extra, winPwd, bobRoot))

        solutionProjectList = []
        solutionProjectConfigs = []
        for name,project in projects.items():
            p = os.path.join(self.destination, name)
            os.makedirs(p, exist_ok=True)
            self.updateFile(os.path.join(p, name+".vcxproj"), project.generateProject(projects, buildMeCmd),
                    encoding="utf-8", newline='\r\n')
            self.updateFile(os.path.join(p, name+".vcxproj.filters"), project.generateFilters(),
                    encoding="utf-8", newline='\r\n')
            if not self.args.update:
                self.updateFile(os.path.join(p, name+".vcxproj.user"), project.generateUser(),
                        encoding="utf-8", newline='\r\n')

            solutionProjectList.append(SOLUTION_PROJECT_TEMPLATE.format(NAME=name, GUID=str(project.uuid).upper()))
            solutionProjectConfigs.append("\t\t{{{GUID}}}.Build|x86.ActiveCfg = Build|Win32".format(GUID=str(project.uuid).upper()))
            solutionProjectConfigs.append("\t\t{{{GUID}}}.Checkout+Build|x86.ActiveCfg = Checkout+Build|Win32".format(GUID=str(project.uuid).upper()))
            if project.isRoot:
                solutionProjectConfigs.append("\t\t{{{GUID}}}.Build|x86.Build.0 = Build|Win32".format(GUID=str(project.uuid).upper()))
                solutionProjectConfigs.append("\t\t{{{GUID}}}.Checkout+Build|x86.Build.0 = Checkout+Build|Win32".format(GUID=str(project.uuid).upper()))

        self.updateFile(os.path.join(self.destination, self.projectName+".sln"),
                SOLUTION_TEMPLATE.format(PROJECTS_LIST="\n".join(solutionProjectList),
                                         PROJECTS_CFG="\n".join(solutionProjectConfigs),
                                         SOLUTION_GUID=str(self.uuid)),
                encoding="utf-8", newline='\r\n')

    def __generateBuildme(self, extra, projectRoot, bobRoot):
        buildMe = []
        if sys.platform == 'msys':
            extra = " ".join(quoteBash(e) for e in extra)
            buildMe.append("#!/bin/sh")
            buildMe.append("export PATH=" + quoteBash(os.environ["PATH"]))
            buildMe.append('bob dev "$@" ' + extra)
            quote = quoteBash
        else:
            extra = " ".join(quoteCmdExe(e) for e in extra)
            buildMe.append("@ECHO OFF")
            buildMe.append("cd " + quoteCmdExe(str(projectRoot)))
            buildMe.append(bobRoot + ' dev %* ' + extra)
            quote = quoteCmdExe
        projectCmd = bobRoot + " project -n " + extra + " vs2019 " + quote("/".join(self.rootPackage.getStack())) + \
            " -u --destination " + quote(self.destination) + ' --name ' + quote(self.projectName) + \
            " --uuid " + quote(str(self.uuid))
        # only add arguments which are relevant for .files or .includes. All other files are only modified if not build with
        # update only.
        for i in self.args.additional_includes:
            projectCmd += " -I " + quote(i)
        if self.args.filter:
            projectCmd += " --filter " + quote(self.args.filter)
        for e in self.args.excludes:
            projectCmd += " --exclude " + quote(e)
        for e in self.args.include:
            projectCmd += "--include " + quote(e)
        for e in self.args.start_includes:
            projectCmd += " -S " + quote(e)

        buildMe.append(projectCmd)
        return "\n".join(buildMe)


def vs2019ProjectGenerator(package, argv, extra, bobRoot):
    generator = Vs2019Generator()
    generator.configure(package, argv)
    generator.generate(extra, bobRoot)
